#
# estcore.rb
#
#     Copyright(c) 1999 Minero Aoki
#     aamine@dp.u-netsurf.ne.jp
#

require 'nkf'
require 'thread'
require 'delegate'
require 'ftools'
require 'singleton'

require 'amstd/must'
require 'amstd/extmod'
require 'amstd/bug'

require 'tmail/tmail'

require 'est/smtpclient'
require 'est/popclient'
require 'est/readrc'


module EST

  Version = '0.0.3'

	def EST.create
		const_set :CORE, Core.instance
	end


  class CCore
	
	  include Singleton

    #
    # export
    #
    #   rc, [],
    #   mailset, composer, sender, popper,
    #   notice, shutdown
    #

    private

    def do_initialize
      init_rc

      @composer = Composer.new
      @sender   = Sender.new
      @popper   = Popper.new
      @mailset  = Mailset.new

			load confdir + '/eval'
    end

    def init_rc
			confdir = File.expand_path( '~/.est' )
			unless File.exist? confdir then
			  File.mkpath confdir
			end
			unless File.directory? confdir then
			  puts "cannot create ~/.est (file exists): remove ~/.est and try again"
				exit 1
			end

      require 'est/estconf'
      @rc = {}
			crc = corerc()
      crc['confdir'] = confdir
      crc['addrfile'] = confdir + '/address'
      @rc['core'] = crc
			@rc['if'] = {}
    end

		def read_config( fn = nil )
			unless fn then
			  fn = @rc['confdir'] + '/config'
			end
      @rc.update setrc( fn )
		end
		public :read_config


    def chk( tag, val )
      # puts "#{tag}: #{val.inspect}"
    end

    def setrc( fname )
      p = ESTrcp.new
      ret = p.parse( fname )

      #-------------------------------------------------------
      # data
      #

      kconfig    = ret['config']
      ksender    = ret['sender']['server']
      kpopper    = ret['popper']['server']
      kloader    = ret['mailbox']['mbox']
      kcomposer  = ret['composer']['entry']
      kinterface = ret['interface'][ EST::Interface ]
      krefile    = ret['mailbox']['refile']

      chk 'kconfig', kconfig
      chk 'ksender', ksender
      chk 'kpopper', kpopper
      chk 'kloader', kloader
      chk 'kcomposer', kcomposer
      chk 'kinteface', kinterface
      chk 'krefile', krefile

      #-------------------------------------------------------
      # profile
      #

      conf = kconfig[ ret[ 'use' ] ]

      vsender = ksender[ conf[ 'sender' ] ]
      vsender['port'] = vsender['port'].to_i
      
      vpopper = {}
      conf[ 'popper' ].split(',').each do |srv|
        srv.strip!
        phash = {}
        kpopper[ srv ].each do |k,v|
          case k
          when 'port'
            v = v.to_i
          end
          phash[ k ] = v
        end
        vpopper[ srv ] = phash
      end
      vloader   = ret[ 'mailbox' ][ 'system' ][ conf[ 'mailbox' ] ]
      vcomposer = kcomposer[ conf[ 'composer' ] ]
      vrefile   = mkrefilearg( krefile )

      chk 'conf', conf
      chk 'vsender',   vsender
      chk 'vpopper',   vpopper
      chk 'vloader',   vloader
      chk 'vcomposer', vcomposer
      chk 'vrefile',   vrefile

      vl = {}
      (kloader.keys - vloader.values).each do |name|
        vl.store name, kloader[name]
      end
      vloader.each do |name,real|
        vl.store name, kloader[real]
      end
      vloader = vl

      #-------------------------------------------------------
      # set
      #

      ret = {
        'current' => {
          'sender' => vsender,
          'popper' => vpopper,
          'mailbox' => vloader,
          'composer' => vcomposer
        },
        'sender' => ksender,
        'popper' => kpopper,
        'mailbox' => kloader,
        'composer' => kcomposer,
        'if' => kinterface
      }

      return ret
    end


    def mkrefilearg( arg )
      return nil unless arg

      arr = []
      arg.each do |num, str|
        # "str" is like [ header, refile_regex, dest_mbox ]
        arr[ num.to_i ] = eval( str )
      end
      arr.compact!

      return arr
    end



    public


    attr :rc
    attr :mailset
    attr :composer
    attr :sender
    attr :popper

    def []( key ) @rc['current'][ key ] end

    def notice( str )
      $stderr.puts str
    end
    
    def shutdown
puts 'EST core.shutdown ...'
      @mailset.shutdown if @mailset
      puts 'done'
    end

  end

  class Core < CCore
	end



  class CMailset

    #
    # export
    #
    #   fetch, [], mbox_names,
    #   mvmail, cpmail, rmmail, mkmail,
    #   refile
    #

    def initialize
      @set = {}
    end

    def shutdown
puts 'mailset shutdown'
      @set.each do |name, mb|
        if Mailbox === mb then
          mb.finalize
        end
      end
    end


    def fetch( realname )
      realname.must String   #D#

      unless ret = @set[ realname ] then
        if rc = CORE['mailbox'][realname] then
          lo = mkloader( rc )
          ret = realizembox( realname, rc, lo )
          @set.store( realname, ret )
        else
          ret = nombox( realname )
        end
      end

      return ret
    end

    alias [] fetch
      

    def mbox_names
      ret = []
      CORE['mailbox'].each do |name, rc|
        if vn = rc['viewname'] then
          ret.push vn
        else
          ret.push name
        end
      end
      ret.sort!

      return ret
    end


    def mkloader( rc )
      case rc[ 'type' ]
      when 'mh'
        lo = MhLoader.new( rc['path'] )
      else
        bug! 'mbox type is not MH'
      end

      return lo
    end

    def realizembox( name, rc, lo )
      Mailbox.new( name, rc, lo )
    end

    def nombox( name )
      raise NameError, "no such mailbox: #{name}"
    end

    def get_targetmbox( name )
      self[ name ]
    end


    def mvmail( from_name, idx, to_name )
      return if from_name == to_name
      return unless from = self[ from_name ]
      return unless to = get_targetmbox( to_name )
      return unless port = from[idx]

      to.mv port
      from.rm idx
    end


    def cpmail( idx, from_name, to_name )
      return unless from = self[ from_name ]
      return unless to = get_targetmbox( to_name )
      return unless port = from[idx]

      to.cp port
    end


    def rmmail( box_name, idx )
      return unless mbox = self[ box_name ]
      mbox.rm idx
    end


    def mkmail( box_name, str = nil )
      return nil unless mbox = self[ box_name ]

      port = mbox.new_port
      port.wopen do |wp|
        if str then
          wp.write str
        else
          yield wp
        end
      end

      mbox.add port

      return port
    end


    def refile
      to_mb = nil
      inbox = self['inbox']

      inbox.each_with_index do |port, idx|
        to_mb = nil
        nil.each do |arr|
          nm, re, to = *arr
          mail = Mail.new( port )

          if mail[nm] =~ re then
            to_mb = to
            break
          end
        end
        mvmail( 'inbox', idx, to_mb ) if to_mb
      end
    end

  end

  class Mailset < CMailset
  end



  class CMailbox

    include Enumerable


    private

    def initialize( name, rc, lo )
      name.must String
      @name = name.dup.freeze
      @rc = rc
      @loader = lo
      @array = []
    end

    def do_initialize
    end

    abstract :do_add
    abstract :do_cp
    abstract :do_mv
    abstract :do_rm
    abstract :do_new_port
    abstract :do_finalize


    public


    def finalize
      do_finalize
    end

    def add( mail, cache = nil )
      do_add mail
    end
        
    def cp( port )
      do_cp port
    end

    def mv( port )
      do_mv port
    end

    def rm( idx )
      do_rm idx
    end

    def new_port
      do_new_port
    end

    
    def update
    end

    attr :name

    def []( idx )
      return nil unless idx
      @array[idx]
    end

    def size
      @array.size
    end
    alias length size

    
    def each( &block )
      @array.each( &block )
    end

    def each_index( &block )
      @array.each_index( &block )
    end

		def each_mail
		  each{|port| yield Mail.new( port ) }
		end

    def index( port )
      port.must Port
      @array.index port
    end

    def rindex( port )
      port.must Port
      @array.rindex port
    end

    def indexes( *args )
      @array.indexes *args
    end

    def sort( &block )
      @array.sort!( &block )
    end

    def to_a
      @array.dup
    end

  end

  class Mailbox < CMailbox
  end



  class RealMailbox < Mailbox

    def update
      hash = {}
      each{|port| hash[ port ] = true }
      @loader.each do |p|
        p.unread = true if @name == 'inbox'
        add p unless hash.has_key? p
      end
    end


    private


    def do_initialize
      if entries = load_cache then
				entries.each do |ent|
				  port = ent.shift
          add port, ent
        end
      else
puts 'parsing'
				unread = (@name == 'inbox')
        @loader.each do |p|
          p.unread = unread
          add p
        end
      end
    end

    def load_cache
      fn = "#{CORE.rc['core']['confdir']}/cache/#{EST::Interface}/#{@name}"
			entries = nil

			begin
				if File.exist? fn then
					File.open( fn ) do |f|
						break if f.gets.downcase.strip != 'est cache file version 1'
puts 'version ok'

						size = f.gets.strip.to_i
						sign = Marshal.load( f.read size )
						break unless correct_cache? sign
puts 'correct cache'

puts "loading cache: #{@name}"
						entries = Marshal.load( f )
					end
				end
unless entries then
puts "cache miss: #{@name}"
end
			rescue
			  $stderr.puts 'error in reading cache file (may not be fatal)'
				$stderr.puts $!.to_s
				entries = nil
      end

			entries
    end

		# abstract correct_cache?()


    def do_finalize
			save_cache if @loader.can_cache?
    end

    def save_cache
puts "saving cache: #{@name}"
      crc = CORE.rc['core']
      fn = "#{crc['confdir']}/cache/#{EST::Interface}/#{@name}"

			begin
				File.mkpath File.dirname(fn)
				File.open( fn, 'w' ) do |f|
					f.puts 'est cache file version 1'

					str = Marshal.dump( cache_signature )
					f.puts str.size.to_s
					f.write str

					Marshal.dump( cache_entries, f )
				end
			rescue
			  $stderr.puts 'error in writing cache file (may not be fatal)'
				$stderr.puts $!.to_s
				if File.file? fn then
				  File.rm_f fn
				end
			end
    end

		# abstract cache_entries()
		# abstract cache_signature()


    def do_add( port )
      port.must Port
      @array.push port
      port.ref

      return port
    end

    def do_cp( port )
      port.must Port

      p = @loader.new_port
      port.cp p
      p.unread = port.unread
      add p

      return p
    end

    def do_mv( port )
      port.must Port

      p = @loader.new_port
# puts "do_mv: from=#{port},to=#{p}"
      port.mv p
      p.unread = port.unread
      add p

      return p
    end

    def do_rm( idx )
      return unless port = @array.delete_at( idx )
      port.unref
      return nil
    end

    def do_new_port
      return @loader.new_port
    end
  
  end


  class ReferenceError < StandardError ; end


  class VirtualMailbox < Mailbox

    private

    def do_finalize
    end

    def do_add( port )
      port.must Port
      @array.push port
      port.ref
    end

    def do_cp( port )
      raise ReferenceError, "cannot copy port to virtual mbox"
    end

    def do_mv( port )
      raise ReferenceError, "cannot move port to virtual mbox"
    end

    def do_rm( idx )
      port = @array.delete_at( idx )
      port.unref if port
    end

    def do_new_port
      raise ReferenceError, "virtual mbox cannot make connection"
    end

  end



  class Mail < TTMail

    include TMailEasyAccess

    def initialize( port )
      port.must Port
      super
    end

    def port
      @port
    end

    def port=( p )
      unless @parsed then
        body
      end
      @port.mv p
      @port = p
    end
      
    def sync
      @port.wopen do |wp|
        wp.write to_s
      end
    end

  end



  class CPopper

    def load
      CORE['popper'].each do |name, rc|
        do_pop( name, rc )
      end
    end

    
    private


    def do_pop( name, rc )
      client = nil
      addr = rc[ 'address' ]
      port = rc[ 'port' ]
      acnt = rc[ 'account' ]
      pass = rc[ 'password' ]
      if_nopassword( acnt, pass ) unless pass

      klass = POP3client
      client = klass.new( addr, port )
      if_loggingin( client )
      begin
        client.login( acnt, pass )
        if_loggedin( client )
      ensure
        client.logout
      end
      if_loggedout( client )
    end

    def if_nopassword( acnt, pass )
      raise ArgumentError, "POP password is not given"
    end

    def if_loggingin( client )
    end

    def if_loggedin( client )
      client.each do |popmail|
        port = CORE.mailset.mkmail( 'inbox' ){|wp|
          popmail.mail wp
        }
        if_popmail port
        popmail.delete
      end
    end

    def if_popmail( port )
      port.unread = true
    end

    def if_loggedout( client )
    end

  end

  class Popper < CPopper
  end



  class CSender

    def sendmail
      if CORE.mailset['draft'].size == 0 then
        if_nosendmail
        return
      end

      client = nil
      arg = CORE['sender']

      addr = arg['address']
      port = arg['port']
      helo = arg['helo-domain']

      klass = SMTPclient
      client = klass.new( addr, port )
      if_loggingin( client )
      begin
        client.login( helo )
        if_loggedin( client )
      ensure
        client.logout
      end
      if_loggedout( client )
    end


    private


    def if_nosendmail
    end

    def if_loggingin( client )
    end

    def if_loggedin( client )
      CORE.mailset['draft'].each_with_index do |port,idx|
        mail = Mail.new( port )
        fr = fromaddr( mail )
        to = toaddrs( mail )
        str = CORE.composer.mailtext( mail, client )
        client.sendmail( str, fr, to )
        if_sendmail( mail, idx )
      end
    end

    def if_loggedout( client )
      return unless tomb = CORE['sender']['moveto']

      (0 .. CORE.mailset['draft'].size - 1).each do |i|
        CORE.mailset.mvmail( 'draft', i, tomb )
      end
    end

    def if_sendmail( mail, idx )
    end


    ###


    def fromaddr( mail )
      faddr = mail.from
      unless faddr then
        raise ArgumentError, "no 'from' address designation in mail"
      end

      faddr
    end

    def toaddrs( mail )
      toaddr = []
      mail.each_dest do |addr|
        toaddr.push addr.addr
      end
      if toaddr.size == 0 then
        raise ArgumentError,
          "no 'to' address designation in mail"
      end

      toaddr
    end

    def mailtext( mail, client )
      mail.body = NKF.nkf( '-j', mail.body )
      mail.to_s( "\r\n" )
    end

  end

  class Sender < CSender
  end



  class CComposer

    def initialize
    end


    REPLYTO_HEADER = {
      'To'          => %|replyto_to(original)|,
      'Subject'     => %|replyto_subject(original)|,
      'In-Reply-To' => %|original.msgid|,
      'References'  => %|original.msgid|
    }

    def replyto( original )
      case original
      when Mail
        ;
      when Port
        original = Mail.new( original )
      else
        original.must Mail, Port
      end

      port = CORE.mailset.mkmail( 'draft' ){|wp|
        NEW_HEADER.dup.update( REPLYTO_HEADER ).each do |n,v|
          wp.puts( n + ': ' + eval(v) )
        end
        wp.puts
        replyto_body( wp, original )
      }
      return port
    end


    def replyto_to( mail )
      if hed = mail['reply-to']        then return hed.body end
      if hed = mail['from']            then return hed.body end
      if hed = mail['resent-reply-to'] then return hed.body end
      if hed = mail['resent-from']     then return hed.body end
      return ''
    end

    
    def replyto_subject( original )
      sbj = original.subject('')
			sbj = 
			  if /\A\s*(?:\[[^\]]+\])?\s*(?:re\:\s*)*(\S.*)\Z/io === sbj then
          'Re: ' + $1
				else
					'Re: ' + sbj
				end

      return sbj
    end

    
    def replyto_body( mail )
      str = <<EOS


In mail "%s"
  %s wrote:

EOS
			str = sprintf( str,
			  mail.subject(''), mail.from_phrase('') || mail.from('') )
      mail.body.each {|line| str << '> ' << line }

      return str
    end


    NEW_HEADER = {
      'Date'         => %|DateH.t2s(Time.now)|,
      'Subject'      => %||,
      'From'         => %|CORE['composer']['from-address']|,
      'MIME-Version' => %|"1.0"|,
      'Content-Type' => %|"text/plain; charset=iso-2022-jp"|,
      'X-Mailer'     => %|'Mailest version ' + EST::Version|
    }

    def new_mail
      port = CORE.mailset.mkmail( 'draft' ){|wp|
        NEW_HEADER.each do |n,v|
          wp << n << ': '
					wp.puts eval(v)
        end
				wp.puts
      }

      return port
    end


		def forward( original )
		  port = CORE.mailset.mkmail( 'draft' ){|wp|
			  NEW_HEADER.each do |n,v|
				  wp << n << ': '
					wp.puts eval(v)
				end
				if hf = original['to'] then
					if s = hf.addrs.join(', ') then
						wp << 'Resent-To: '
						wp.puts s
					end
				end
				if hf = original['from'] then
				  if s = hf.addrs.join(', ') then
					  wp << 'Resent-From: '
						wp.puts s
					end
				end
				wp.puts
				wp.puts '---------- forwarded mail:'
				original.body.each do |line|
				  wp.puts line
				end
				wp.puts '---------- end forwarded mail'
			}

			return port
		end


    def list_items( m )
			[ list_date( m['date'] && m['date'].date ),
			  list_from( m ),
			  list_subject( m.subject ) ]
    end

    def list_date( ti )
      if ti
        ti.strftime( '%m/%d %H:%M' )
      else
        ''
      end
    end

    def list_from( tm )
      tm.from_phrase(nil) || tm.from('')
    end

    def list_subject( str )
			if /\A\s*\[[^\]]+\]\s*/o === str then
        $'
      else
        str
      end
    end


    VIEW_HEADER = %w( date from subject x-mailer )

    def view_upper( mail )
      arr = []
      mail.each do |k,v|
				arr.push v.inspect unless VIEW_HEADER.include? k
      end
      return arr.join("\n")
    end


    def view_lower( mail )
      tmp = VIEW_HEADER.collect{|k| mail[k].inspect if mail[k] }
      return tmp.compact.join("\n") << "\n\n" << mail_to_viewtext( mail ) 
    end


    def mail_to_viewtext( mail )
      ret = nil
      mtype = mail.maintype.downcase

      case mtype
      when 'text', ''
        ret = mail.body
        case mail.encoding('').downcase
        when '7bit', '8bit', ''
          ;
        when 'base64'
          ret = ret.unpack('m')[0]
        else
          return undecodable_mail( mail )
        end
      when 'multipart'
        mail.parts.each do |m|
          tmp = m.maintype.downcase
          if tmp == 'text' or tmp == '' then
            return mail_to_viewtext(m)
          end
        end
        return unknown_type( mail )
      else
        return unknown_type( mail )
      end

      if NKF.guess( ret ) != NKF::EUC then
        ret = NKF.nkf( '-e', ret )
      end

      # search for "\r\n" from first 200 char
      from = ret.size - 200
      from = from > 0 ? from : 0
      if ret.index( "\r", from ) then
        ret = ret.gsub( "\r\n", "\n" )
      end

      return ret
    end

    def undecodable_mail( mail )
			"cannot decode this mail\n" +
      mail['content-transfer-encoding'].inspect
    end

    def unknown_type( mail )
			"unknown content type\n" +
      mail['content-type'].inspect
    end

    def standby_send( mail )
      mail.to_s
    end

  end

  class Composer < CComposer
  end


end   # module EST



# post-define Port (in 'port.rb')

class Port

  def initialize
    @refc = 0
    @unread = false
  end

  attr :unread, true

  def ref
    @refc += 1
  end

  def unref
    @refc -= 1
#puts "port#unref, refc=#{@refc}"
    if @refc < 1 then
#puts "delete file"
      self.rm
    else
#puts "not delete"
    end
  end

end



# post-defines

class FilePort

  def _dump( limit )
    @filename
  end

  def FilePort._load( str )
    new( str )
  end

end

class Loader

  def can_cache?
    false
  end

end

class MhLoader

  def can_cache?
    true
  end

end
